/* $Module: Redirect input file               $ */
/* $Author: Agma Traina and Caetano Traina Jr $ */
/* $Date: July 5, 1999                        $ */
/*----------------------------------------------*/

#include "CommandProc.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

// ---- GNU GCC ----
#ifdef __GNUC__
  #include <unistd.h>
#endif

//---------------------------------------------------------------------------
//  Support to redirect input
#define STDIN 0
#define COMMPROC_MAX_INPUT_FILES  10

int CurrentInputFileId = 0;
int ContLine=0;
FILE *InputFiles[COMMPROC_MAX_INPUT_FILES] = {stdin};

char *variable[10][10]={{0,0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0},
                        {0,0,0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0,0,0}};
#define CMMPROC_LINESIZE 1024
char line[CMMPROC_LINESIZE];
char CurrentCommandFileName[200]={0};
size_t iline= CMMPROC_LINESIZE + 1;
int level_limit=0;

//---------------------------------------------------------------------------
void InpFile_SetLimit(int i) { level_limit=i; }

//---------------------------------------------------------------------------
int redirect (char *fname) { // redirect setdin to given file
  FILE *fh;
    if (CurrentInputFileId >= COMMPROC_MAX_INPUT_FILES) return -1;
    fh=fopen (fname, "r");    //open temporary file
    if (fh== NULL) return -2;

    //create a duplicate handle for standard input
    InputFiles[++CurrentInputFileId] = fh;
    strcpy(CurrentCommandFileName, fname);
    return 0;
 }

 //---------------------------------------------------------------------------
int dedirect () { // restore the old input file
    /* restore original standard input handle */
    if (CurrentInputFileId <= 0) return -1;
    fclose(InputFiles[CurrentInputFileId]);
    InputFiles[CurrentInputFileId--] = NULL;
    return 0;
}

//---------------------------------------------------------------------------
int isdirected () { // return true if there are a file redirected
    if (CurrentInputFileId <= 0) return false;
    return true;
}

//---------------------------------------------------------------------------
void getline(char *Question) {
  int len;

  if (CurrentInputFileId <= 0) printf ("%s", Question);
  fgets(line, CMMPROC_LINESIZE, InputFiles[CurrentInputFileId]);
  len = strlen(line);
  // the checking for \r is due to problems with transfering
  // files between Linux and Windows.
  if ((line[len - 1] == '\n') || (line[len - 1] == '\r'))
    line[len - 1] = '\0'; // to replace new line char '\n'
  if (line[len - 2] == '\r')
    line[len - 2] = '\0'; // to replace new line char '\n'
  iline=0;
}

//---------------------------------------------------------------------------
int ReadWord (char *cmdLine, int lower, char *Question) {
/* Read the next word.
 *    Strip off the comments ( characters following //), expand macros.
 *    Return:
 *        cmdLine - the word read
 *        return value - 1 if a new line was read, 0 otherwise
 *    Inputs:
 *        lower - enables lowercase substitution
 *        Question - String printed to user if a new line need to be read
 */
  int n, i, j;
  size_t k;
  i = j = 0;
  if (iline > CMMPROC_LINESIZE || iline > strlen(line)) {
    getline(Question);
    if (ContLine) {
      ContLine = 0;
      j = 0;
    }
    else
      j = 1;
  }

  while (1) {
    if ((line[iline] == '/') && (line[iline+1]== '/') ) {
      if (i>0)
        break;
      getline(Question);
      if (ContLine) {
        ContLine = 0;
        j = 0;
      }
      else
        j = 1;
      continue;
    }
    if (line[iline]== '%' && (line[++iline]>='0' && line[iline]<='9') ) {
      n = line[iline]-'0';
      if (variable[CurrentInputFileId][n] == 0)
        continue;
      for (k = 0; k < strlen(variable[CurrentInputFileId][n]); k++)
          cmdLine[i++] = variable[CurrentInputFileId][n][k];
      iline++;
      continue;
    }
    if (lower)
      cmdLine[i] = (char)tolower(line[iline++]);
    else
      cmdLine[i] = line[iline++];
    if (cmdLine[i] == ' ' || cmdLine[i] == ',' || cmdLine[i] == ';' ||
        cmdLine[i] == 0) {
      if (i > 0) {
        cmdLine[i] = 0;
        break;
      }
    }
    else if (cmdLine[i]=='\\') {
      if (iline < strlen(line))
        cmdLine[i++] = line[iline++];
      else
        ContLine=1;
    }
    else i++;
    if (iline >= strlen(line) ) {
      if (i > 0) {
        cmdLine[i] = 0;
        break;
      }
      getline(Question);
      if (ContLine) {
        ContLine = 0;
        j = 0;
      }
      else
        j = 1;
    }
  }
  return j;
}

//---------------------------------------------------------------------------
void ReadCommand (char *cmdLine, int lower, char *Question) {
/* Read the next word.
 *    Strip off the comments ( characters following //), expand macros,
 *    and process the words Read/Open and Quit/Close to redirect/restore
 *    the input files.
 *    Return:
 *        cmdLine - the word read
 *    Inputs:
 *        lower - enables lowercase substitution
 *        Question - String printed to user if a new line need to be read
 */
  int cont, i, n, res;
    while (1) {
        ReadWord(cmdLine, 1, Question);
        if (!strcmp(cmdLine,"close")) {
            if (dedirect()) return;
            if (level_limit > CurrentInputFileId) return;
                continue;
        } else
        if ( (!strcmp(cmdLine, "quit")) || (!strcmp(cmdLine,"abort")) ){
            while (!dedirect());
            return;
        } else
        if ( (!strcmp(cmdLine, "open")) || (!strcmp(cmdLine, "read")) ){
            ReadWord(cmdLine, 0, "File Name ?");
            n=strlen(cmdLine);
            res=redirect(cmdLine);
            if (res==-2) {
                for (i=0; i<n; i++) if(cmdLine[i]=='.') break;
                if (i==n) {cmdLine[i++]='.'; cmdLine[i++]='s';
                           cmdLine[i++]='t'; cmdLine[i++]='c'; cmdLine[i]=0; }
                res=redirect(cmdLine);
                if (res==0) printf ("Openning file: %s\n", cmdLine);
                if (res==-2) {
                  char aux[150];
                    aux[0]='s'; aux[1]='t'; aux[2]='c'; aux[3]='\\';
                    i=0;
                    while ( (aux[i+4]=cmdLine[i]) != 0) i++;
                    res=redirect(aux);
                    if (res==0) printf ("Openning file: %s!\n", aux);
                }
            }
            if (res==-1) {
                printf ("Too many input file redirections!\n");
                continue;
                }
            if (res==-2) {
                printf ("File Not found: %s\n", cmdLine);
                continue;
                }
            i=0;
            cont=ReadWord(cmdLine, 0, 0);
            while (cont==0 && i<10) {
              if (variable[CurrentInputFileId][i]!= 0)
                free (variable[CurrentInputFileId][i]);
              variable[CurrentInputFileId][i]=(char *)malloc(strlen(cmdLine)+2);
              memcpy(variable[CurrentInputFileId][i++],
                                                    cmdLine, strlen(cmdLine)+1);
              cont=ReadWord(cmdLine, 0, 0);
            }
        }
        if (lower) for (i=0; i<(int)strlen(cmdLine); i++)
            cmdLine[i] =(char)tolower(cmdLine[i]);
        break;
    }
}

//---------------------------------------------------------------------------

