//==============================================================================
//  Project: SIRIUS/INFORMEDIA
//  Sub-Project: Omni-Seq Tree
//  File: TStringObj.h
//  Classe: TStringObj
//  Authors: Roberto Figueira Santos Filho
//           Agma Jucy Machado Traina
//           Caetano Traina Junior
//           Christos Faloutsos
//  Date: 06/14/2000
//  Version: 0.0
//
//  Description:
//	  Object to be indexed by vector and metric space.
//    String version.
//==============================================================================

#ifndef StringObjH
#define StringObjH

#include <stdio.h>

//------------------------------------------------------------------------------
// Class TStringObj definitions
//------------------------------------------------------------------------------
class TStringObj {
  public:
    TStringObj(); // null constructor
    TStringObj(int sSize); // main constructor
    TStringObj(char *string, int sSize); // constructs and initialize
    TStringObj(int sSize, char *buff); // mapping constructor
         // It is faster to be used when it is not desired to
         // create a new object. The idea is to share the data space and use the
         // internal methods. It is used to avoid expensive buffer copies.
         // The user must take care about allocating and deleting the object
         // data space. ObjSize can be obtained with another object created by
         // the main constructor.
    TStringObj(const TStringObj&cObj); // copy constructor
    TStringObj(char *fName, int sSize, int nop); // SampleSet constructor
                                    // (to create a query object of type biased)
    TStringObj(int sSize, int value);
                    // randomObject constructor (to create a query object of
                    // type random)
    ~TStringObj();

    // to be used with the null constructor. The obj is used as an example to
    // get definitions (mySize in this case).
    void allocateAs(TStringObj *obj);
    void setSize(int sSize);
    char& operator [](int i);
    const char& operator [](int i) const;
    void setReadObjectFunction(int (*rdFunc)(char *, int, char *));
    void readObj(char *cmdLine, int condition, char *aux);
    void randomize(); // get biased
    void printString();
    void readFromTxtFile(FILE *fdIn);
    void writeToTxtFile(FILE *fdOut);
    TStringObj& operator=(const TStringObj &cObj);
    void toObject(char *buff); // copies the buffer to the object. The buffer
          // MUST have the same size of the object buffer data space
    void toBuffer(char *buff); // copies the object buffer space to the buffer.
          // The user must allocate and delete the buffer
    void mapBuffer(char *buff); // map method - must be used with map
                                // constructor. The sSize must be initialized
                                // with the map constructor. Use NULL as the
                                // buffer parameter if you need.
    char* bufferArea(); // returns a pointer to a buffer area
            // containing the object. The user can NEVER delete this area
    int objectSize() { return(mySize); };

    typedef int distanceType;
    distanceType distance(TStringObj *obj);
    // check if obj is inside the circle (*this + radius). Returns true if it
    // is (and its distance), false otherwise (and 0 as the distance).
    char isObjectIncluded(TStringObj *obj, distanceType radius,
                                               distanceType *dist);
  private:
    enum tObjectClass {
      ocAllocated,
      ocMaped,
      ocSample,
      ocRandom,
      ocNull
    }; // indicates whether the constructor must delete data or not
    tObjectClass myClass;

    int mySize;
    char *data;

    char **sampleSet; // objects to perform biased queries
    int sampleSetSize; // indicates how many objects are in vRand
    int currentSample; // indicates how many objects are in vRand
    int maxValue; // used with random not biased

    int (*readFunc)(char *, int, char *);

    int min3(int a, int b, int c); // Calculates the minimum between 3 int
    distanceType wordDiff(char *s1, char *s2); // Original LEdit metric

    void initAll(); // initialize the object with all NULL
};
//---------------------------------------------------------------------------
#endif
