//==============================================================================
//  Projeto: SIRIUS
//  Arquivo: TNSCache.CPP
//  Classe: TNSCache
//  Autor: Roberto Figueira Santos Filho
//  Orientador: Caetano Traina Junior
//  Data Inicial: 10/06/1997
//  Versao Atual: 0.0
//
//  Descricao:
//	  Este arquivo apresenta os metodos da classe Gerenciador de Memoria Cache,
//  o qual sera usado no Gerenciador do Arquivo Fisico do SIRIUS.
//
//	Observacoes:
//
//  Modificacoes:
//==============================================================================

#include <stdio.h>
//---------------------------------------------------------------------------
#pragma hdrstop

#include "TNSCache.h"

// Declaracao dos metodos da classe TNSCache

//==============================================================================
// Metodo: TNSCache(TpObjCntin *oAux,
//                 void (TpObjCntin::*fLe)(unsigned long, unsigned char *),
//				   void (TpObjCntin::*fEscreve)(unsigned long, unsigned char *),
//                 unsigned long NRegs, unsigned char FRot, unsigned char NPgs,
//				   unsigned long TamBuff)
//==============================================================================
template <class TpObjCntin>
TNSCache<TpObjCntin>::TNSCache(TpObjCntin *oAux,
       void (TpObjCntin::*fLe)(long, void *),
       void (TpObjCntin::*fEscreve)(long, void *),
       unsigned long NRegs, unsigned char FRot,
       unsigned int NPgs, unsigned long TamBuff) {

  register unsigned long i, j;
  register int k;

  // inicializando parametros de configuracao do cache - InicPar()
  F_ROTACAO = FRot;
  RGS_POR_BLOCO = (unsigned char)(1 << F_ROTACAO);
  N_PAGINAS = NPgs;

  // inicializando TME
  TME.TotBlocos = (NRegs >> F_ROTACAO) + TxCrescBlocos;
  TME.pBloco = new unsigned int[TME.TotBlocos];
  for (i = 0; i < TME.TotBlocos; TME.pBloco[i++] = NIL);
  TME.Bloco = new tBloco[N_PAGINAS];
  for (i = 0; i < N_PAGINAS; i++){
    TME.Bloco[i].pPagina = new unsigned int[RGS_POR_BLOCO];
    for (k = 1; k < RGS_POR_BLOCO; TME.Bloco[i].pPagina[k++] = NIL);
    TME.Bloco[i].RgsUsados = 0;
    TME.Bloco[i].pPagina[0] = i + 1; // se bloco nao usado a posicao 0 aponta
    						 // para o proximo disponivel - economiza memoria
  } // fim do for
  TME.Bloco[N_PAGINAS - 1].pPagina[0] = NIL;
  TME.LBlocoDispo = 0;

  // inicializando ListaPgs
  ListaPgs.Pagina = new tPagina[N_PAGINAS];
  for (i = 0; i < N_PAGINAS; i++){
    unsigned char *paux;

    ListaPgs.Pagina[i].proxDaFila = i + 1;
    ListaPgs.Pagina[i].antDaFila = NIL;
    ListaPgs.Pagina[i].IdxTotAcessos = NIL;
    ListaPgs.Pagina[i].IdRegFisico = 0xFFFFFFFF;
    ListaPgs.Pagina[i].Escrita = 0;
    ListaPgs.Pagina[i].Buffer = new unsigned char  [TamBuff];
    paux = ListaPgs.Pagina[i].Buffer;
    for (j = 0; j < TamBuff; *paux++ = '\0', j++);
  } // fim do for
  ListaPgs.Pagina[N_PAGINAS - 1].proxDaFila = NIL;
  ListaPgs.LPgDispo = 0;
  ListaPgs.PgEmUso = NIL;

  // inicializando GerAcessos
  GerAcessos.LTotAcessos = new tLTotAcessos[N_PAGINAS];
  for (i = 0; i < N_PAGINAS; i++){
    GerAcessos.LTotAcessos[i].TotAcessos = 0;
    GerAcessos.LTotAcessos[i].prox = i + 1;
    GerAcessos.LTotAcessos[i].ant = NIL;
    GerAcessos.LTotAcessos[i].frontFilaPg =
    		   GerAcessos.LTotAcessos[i].rearFilaPg = NIL;
  } // fim do for
  GerAcessos.LTotAcessos[N_PAGINAS - 1].prox = NIL;
  GerAcessos.Cabeca = NIL;
  GerAcessos.LDispo = 0;

  // inicializando parametros restantes
  ObjContinente = oAux;
  leC = fLe;
  escreveC = fEscreve;
} // fim do construtor TNSCache

//==============================================================================
// Metodo: ~TNSCache()
//==============================================================================
template <class TpObjCntin>
TNSCache<TpObjCntin>::~TNSCache() {
  unsigned int i;

  flushCache();

  // liberando TME
  for (i = 0; i < N_PAGINAS; delete[] TME.Bloco[i++].pPagina);
  delete[] TME.Bloco;
  delete[] TME.pBloco;

  // liberando ListaPgs
  for (i = 0; i < N_PAGINAS; delete[] ListaPgs.Pagina[i++].Buffer);
  delete[] ListaPgs.Pagina;

  // liberando GerAcessos
  delete[] GerAcessos.LTotAcessos;
} // fim do destrutor ~TNSCache

//==============================================================================
// Metodo: void GMC_le(pal4, void **)
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::readBuffer(unsigned long reg, void **buffer) {
  register unsigned long bloco = (reg >> F_ROTACAO); // gasta em torno
  									// de 1 ciclo dependendo da arquitetura
  register unsigned int pBloco = TME.pBloco[bloco];
  register unsigned int bit = (unsigned int)(reg - (bloco << F_ROTACAO));
  					     // gasta em torno de 2 ciclos dependendo da arquitetura
  register unsigned int pPg = ((pBloco != NIL) ?
                               (TME.Bloco[pBloco].pPagina[bit]) : NIL);

  if (pPg != NIL){ // a pagina ja esta na memoria
    ListaPgs.PgEmUso = pPg;
    IncPagina(); // incrementa a pagina em uso
    *buffer = (void *)ListaPgs.Pagina[pPg].Buffer;
    return;
  } // fim do then
  else { // o registro nao esta na memoria
    pPg = ObtemPagina();
    // obtem registro da base
	(ObjContinente->*leC)(reg, ListaPgs.Pagina[pPg].Buffer);
    ListaPgs.PgEmUso = pPg;
	ListaPgs.Pagina[pPg].IdRegFisico = reg;

//    IncPagina(); // incrementa a pagina em uso mesmo que seja nova, pois evita
    			 // que sempre as mesmas paginas sejam liberadas
    *buffer = (void *)ListaPgs.Pagina[pPg].Buffer;

    // atualiza TME
    // IMPORTANTE: o teste abaixo nao pode ser "if (pBloco == NIL)"!
    // Observe que se TME.pBloco[bloco] possuir no minimo um registro, pBloco
    // e diferente de NIL. Porem, se a pagina obtida coincidir com o registro
    // de TME.pBloco[bloco] (que so contem este registro), TME.pBloco[bloco]
    // passa a ser NILL (ver // IMPORTANTE REF em ObtemPagina()), gerando erro
    // com relacao a pBloco que antes era diferente de NIL, visto que existia
    // um registro neste bloco.
    if (TME.pBloco[bloco] == NIL){ // nao existe nenhum registro deste bloco
    							   // na memoria
      // obter um bloco livre e incluir no ponteiro de blocos
      pBloco = TME.LBlocoDispo;
      // o teste abaixo pode ser retirado, se o cache estiver funcionando
      // corretamentepara aumentar desempenho
      if (pBloco == NIL) { //nunca deve acontecer - erro de logica
        printf("TNSCache::GMC_le - Erro ao tentar obter um bloco");
        printf(" livre!\n");
        exit(1);
      } // fim do if
      TME.LBlocoDispo = TME.Bloco[pBloco].pPagina[0];
      TME.Bloco[pBloco].pPagina[0] = NIL;
      TME.Bloco[pBloco].RgsUsados = 0;
      TME.pBloco[bloco] = pBloco;
    } // fim do if
    TME.Bloco[pBloco].pPagina[bit] = pPg;
    TME.Bloco[pBloco].RgsUsados++;
  } // fim do if
} // fim do metodo GMC_le

//==============================================================================
// Metodo: void newBuffer(unsigned long reg, void **buffer)
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::newBuffer(unsigned long reg, void **buffer) {
  if (reg >=  (TME.TotBlocos * RGS_POR_BLOCO)) { // estouro do numero de blocos
    unsigned long novoTotal = TME.TotBlocos + TxCrescBlocos;
    register unsigned long i;

    TME.pBloco = (unsigned int *)realloc(TME.pBloco,
                                              sizeof(unsigned int) * novoTotal);
    for (i = TME.TotBlocos; i < novoTotal; TME.pBloco[i++] = NIL);
    TME.TotBlocos = novoTotal;
  } // fim do if
  readBuffer(reg, buffer);
} // fim do metodo GMC_NovoNRgs

//==============================================================================
// Metodo: void flushCache()
// Descarrega as paginas alteradas em disco
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::flushCache() {
  register unsigned int i, j;

  i = GerAcessos.Cabeca;
  while (i != NIL) {
    j = GerAcessos.LTotAcessos[i].rearFilaPg;
    while (j != NIL) {
      if (ListaPgs.Pagina[j].Escrita)
        (ObjContinente->*escreveC)(ListaPgs.Pagina[j].IdRegFisico,
        						   ListaPgs.Pagina[j].Buffer);
      // fim do if
      j = ListaPgs.Pagina[j].proxDaFila;
    } // fim do while
    i = GerAcessos.LTotAcessos[i].prox;
  } // fim do while
} // fim do metodo GMC_flushCache

//==============================================================================
// Metodo: void IncPagina()
// incrementa o numero de acessos da pagina corrente
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::IncPagina() {
  register unsigned int pPgEmUso = ListaPgs.PgEmUso;
  register unsigned int pTotAces = ListaPgs.Pagina[pPgEmUso].IdxTotAcessos;
  register unsigned int pProxTotAces = GerAcessos.LTotAcessos[pTotAces].prox;

  if (pProxTotAces != NIL) { // este nao e o ultimo TotAcessos
    if ((GerAcessos.LTotAcessos[pTotAces].TotAcessos + 1) ==
    	GerAcessos.LTotAcessos[pProxTotAces].TotAcessos) {
            // ProxTotAcessos e um a mais que o atual
      TiraPgFila(pPgEmUso, pTotAces);
      ColocaPgFila(pPgEmUso, pProxTotAces);
    } // fim do then
    else { // ProxTotAcessos e mais do que um a mais que o atual
      if (GerAcessos.LTotAcessos[pTotAces].frontFilaPg ==
          GerAcessos.LTotAcessos[pTotAces].rearFilaPg) {
          // esta e a unica pagina neste TotAcessos e o total seguinte e
          // maior do que este incrementado, basta incrementar o total
        GerAcessos.LTotAcessos[pTotAces].TotAcessos++;
      } // fim do then
      else { // existem outras paginas na fila deste TotAcessos e o total
          // seguinte e maior do que este incrementado. Um novo total deve ser
          // criado e esta pagina deve ser transferida para ele.
        TiraPgFila(pPgEmUso, pTotAces);
        NovoTotal(pPgEmUso, pTotAces); // pTotAces e alterado retornando o novo
      } // fim do if
      if (GerAcessos.LTotAcessos[pTotAces].TotAcessos > 0xFFEF) {
          // esta proximo do valor maximo, todos os totais sao reajustados
          // a comparacao foi feita com 0xFFEF e nao com 0xFFFF por seguranca
        register unsigned int pAux = GerAcessos.Cabeca;
        register unsigned int novoTot = 1;

        while (pAux != NIL) {
          GerAcessos.LTotAcessos[pAux].TotAcessos = novoTot++;
          pAux = GerAcessos.LTotAcessos[pAux].prox;
        } // fim do while
      } //fim do if
    } // fim do if
  } // fim do then
  else{ // este e o ultimo TotAcessos
    if (GerAcessos.LTotAcessos[pTotAces].frontFilaPg ==
        GerAcessos.LTotAcessos[pTotAces].rearFilaPg) {
        // esta e a unica pagina neste TotAcessos e o total seguinte e
        // maior do que este incrementado, basta incrementar o total
      GerAcessos.LTotAcessos[pTotAces].TotAcessos++;
    } // fim do then
    else { // existem outras paginas na fila deste TotAcessos e o total
        // seguinte e maior do que este incrementado. Um novo total deve ser
        // criado e esta pagina deve ser transferida para ele.
      TiraPgFila(pPgEmUso, pTotAces);
      NovoTotal(pPgEmUso, pTotAces); // pTotAces e alterado retornando o novo
    } // fim do if
    if (GerAcessos.LTotAcessos[pTotAces].TotAcessos > 0xFFEF) {
    	// esta proximo do valor maximo, todos os totais sao reajustados
        // a comparacao foi feita com 0xFFEF e nao com 0xFFFF por seguranca
      register unsigned int pAux = GerAcessos.Cabeca;
      register unsigned int novoTot = 1;

      while (pAux != NIL) {
        GerAcessos.LTotAcessos[pAux].TotAcessos = novoTot++;
        pAux = GerAcessos.LTotAcessos[pAux].prox;
      } // fim do while
    } //fim do if
  } // fim do if
} // fim do metodo IncPagina()

//==============================================================================
// Metodo: unsigned int ObtemPagina()
//==============================================================================
template <class TpObjCntin>
unsigned int TNSCache<TpObjCntin>::ObtemPagina() {
  register unsigned int pPgLivre = ListaPgs.LPgDispo;
  register unsigned int pTotAces;

  if (pPgLivre != NIL){ // ainda existe uma pagina livre no cache
    ListaPgs.LPgDispo = ListaPgs.Pagina[pPgLivre].proxDaFila;
    ListaPgs.Pagina[pPgLivre].proxDaFila = NIL;
    if (GerAcessos.Cabeca == NIL) { // este e o primeiro total a ser incluido
      pTotAces = GerAcessos.LDispo;
      GerAcessos.LDispo = GerAcessos.LTotAcessos[pTotAces].prox;
      GerAcessos.LTotAcessos[pTotAces].prox = NIL;
      GerAcessos.Cabeca = pTotAces;
      GerAcessos.LTotAcessos[pTotAces].frontFilaPg =
                 GerAcessos.LTotAcessos[pTotAces].rearFilaPg = pPgLivre;
      // GerAcessos.LTotAcessos[pTotAces].TotAcessos ja vale 0,
      // nao perder tempo inicializando aqui
      ListaPgs.Pagina[pPgLivre].IdxTotAcessos = pTotAces;
    } // fim do then
    else { // ja existe pelo menos um TotAcessos, a pPgLivre sera incluida
    	   // na fila de paginas do menor TotAcessos (GerAcessos.Cabeca) e nao
           // em um TotAcessos com 0 para que nao haja uma distancia muito
           // grande entre o atual e o proximo. A inclusao na fila e sempre
           // no GerAcessos.LTotAcessos[pTotAces].rearFilaPg
      ColocaPgFila(pPgLivre, GerAcessos.Cabeca); 
    } // fim do if
    return (pPgLivre);
  } // fim do then
  else { // nao existe pagina livre no cache. Retira pagina acessada a mais
         // tempo da fila de paginas que possui menor frequencia de acessos
    register unsigned long reg;
    register unsigned long bloco;
    register unsigned int pBloco;
    register unsigned int bit;
    register unsigned int pPgAnt, pPgRear, pPgFrontProxTot;
    register unsigned int pProxTotAces;

    pTotAces = GerAcessos.Cabeca;
    pPgLivre = GerAcessos.LTotAcessos[pTotAces].frontFilaPg;
    if (ListaPgs.Pagina[pPgLivre].Escrita) {
      (ObjContinente->*escreveC)(ListaPgs.Pagina[pPgLivre].IdRegFisico,
        						 ListaPgs.Pagina[pPgLivre].Buffer);
      ListaPgs.Pagina[pPgLivre].Escrita = 0;
    } // fim do if
    reg = ListaPgs.Pagina[pPgLivre].IdRegFisico;
    bloco = (reg >> F_ROTACAO);
	pBloco = TME.pBloco[bloco];
    bit = (unsigned int)(reg - (bloco << F_ROTACAO));
    TME.Bloco[pBloco].pPagina[bit] = NIL;
    TME.Bloco[pBloco].RgsUsados--;
    // IMPORTANTE REF : referencia feita por GCM_le sobre tornar
    // TME.pBloco[bloco] = NIL quando o bloco da pagina que foi retirada
    // coincide com o bloco da pagina a ser inserida (ver GCM_le).
    if (!TME.Bloco[pBloco].RgsUsados) { // TME.Bloco[pBloco].RgsUsados = 0,
    					// o bloco pBloco nao possui mais nenhum registro em
                        // memoria. Deve ser liberado entao.
      TME.Bloco[pBloco].pPagina[0] = TME.LBlocoDispo;//inclui na lista de dispo
      TME.LBlocoDispo = pBloco;
      TME.pBloco[bloco] = NIL;
    } // fim do if
    // tira o no da frente da fila e coloca no fim, no mesmo total.
    // primeiro verifica se so existe uma pg na fila: se so for 1 pg, nao faz
    // nada; se forem mais de 1 pgs, tira a da frente e reinsere no fim da fila.
    if (GerAcessos.LTotAcessos[pTotAces].frontFilaPg !=
        GerAcessos.LTotAcessos[pTotAces].rearFilaPg) {//existe + de 1 pg na fila
      pPgAnt = ListaPgs.Pagina[pPgLivre].antDaFila;
      GerAcessos.LTotAcessos[pTotAces].frontFilaPg = pPgAnt;
      ListaPgs.Pagina[pPgAnt].proxDaFila =
    		   ListaPgs.Pagina[pPgLivre].antDaFila = NIL;
      ListaPgs.Pagina[pPgLivre].proxDaFila = pPgRear =
    		   GerAcessos.LTotAcessos[pTotAces].rearFilaPg;
      GerAcessos.LTotAcessos[pTotAces].rearFilaPg =
    		     ListaPgs.Pagina[pPgRear].antDaFila = pPgLivre;
    }
    // incrementa o total desta fila para nao obter uma retirada viciosa.
    // Se o total seguinte for o atual + 1, a fila atual sera incluida
    // na frente da outra, mantendo a ordem em que as paginas seriam retiradas,
    // mas acrescentando as outras paginas para nao obter uma retirada viciosa.
    // Caso contrario, o total atual e incrementado de um.
    pProxTotAces = GerAcessos.LTotAcessos[pTotAces].prox;
    if (pProxTotAces != NIL) { // este nao e o ultimo TotAcessos
      if ((GerAcessos.LTotAcessos[pTotAces].TotAcessos + 1) ==
          GerAcessos.LTotAcessos[pProxTotAces].TotAcessos) {
            // ProxTotAcessos e um a mais que o atual, a fila atual e incluida
            // no final da proxima fila, e o total atual sera disponibilizado
        register unsigned int pPgAux = pPgLivre;

        while (pPgAux != NIL) { // atualiza o indice de totais das paginas da
                                // fila atual para apontar para o novo indice
          ListaPgs.Pagina[pPgAux].IdxTotAcessos = pProxTotAces;
          pPgAux = ListaPgs.Pagina[pPgAux].proxDaFila;
        } // fim do while
        pPgFrontProxTot = GerAcessos.LTotAcessos[pProxTotAces].frontFilaPg;
        ListaPgs.Pagina[pPgLivre].antDaFila = pPgFrontProxTot;
        ListaPgs.Pagina[pPgFrontProxTot].proxDaFila = pPgLivre;
        GerAcessos.LTotAcessos[pProxTotAces].frontFilaPg =
                    GerAcessos.LTotAcessos[pTotAces].frontFilaPg;
        GerAcessos.LTotAcessos[pTotAces].rearFilaPg = NIL;
        GerAcessos.LTotAcessos[pTotAces].frontFilaPg = NIL;
        GerAcessos.Cabeca = pProxTotAces;
        GerAcessos.LTotAcessos[pProxTotAces].ant = NIL;
        GerAcessos.LTotAcessos[pTotAces].TotAcessos = 0;
        GerAcessos.LTotAcessos[pTotAces].prox = GerAcessos.LDispo;
        GerAcessos.LDispo = pTotAces;
      } // fim do then
      else { // ProxTotAcessos e mais do que um a mais que o atual
             // basta incrementar o total
        GerAcessos.LTotAcessos[pTotAces].TotAcessos++;
        if (GerAcessos.LTotAcessos[pTotAces].TotAcessos > 0xFFEF) {
            // esta proximo do valor maximo, todos os totais sao reajustados
            // a comparacao foi feita com 0xFFEF e nao com 0xFFFF por seguranca
          register unsigned int pAux = GerAcessos.Cabeca;
          register unsigned int novoTot = 1;

          while (pAux != NIL) {
            GerAcessos.LTotAcessos[pAux].TotAcessos = novoTot++;
            pAux = GerAcessos.LTotAcessos[pAux].prox;
          } // fim do while
        } //fim do if
      } // fim do if
    } // fim do then
    else{ // este e o ultimo TotAcessos, basta incrementar o total
      GerAcessos.LTotAcessos[pTotAces].TotAcessos++;
      if (GerAcessos.LTotAcessos[pTotAces].TotAcessos > 0xFFEF) {
          // esta proximo do valor maximo, todos os totais sao reajustados
          // a comparacao foi feita com 0xFFEF e nao com 0xFFFF por seguranca
        register unsigned int pAux = GerAcessos.Cabeca;
        register unsigned int novoTot = 1;

        while (pAux != NIL) {
          GerAcessos.LTotAcessos[pAux].TotAcessos = novoTot++;
          pAux = GerAcessos.LTotAcessos[pAux].prox;
        } // fim do while
      } //fim do if
    } // fim do if
    return (pPgLivre); // o total de acessos a essa pg sera incrementado depois
  } // fim do if
}

//==============================================================================
// Metodo: void TiraPgFila(unsigned int pPg, unsigned int pTotAces)
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>:: TiraPgFila(unsigned int pPg, unsigned int pTotAces) {
  register unsigned int pPgAux;
  register unsigned int pTotAcesAux;

  if (GerAcessos.LTotAcessos[pTotAces].frontFilaPg ==
  	  GerAcessos.LTotAcessos[pTotAces].rearFilaPg) {//so existe essa pg na fila
    pTotAcesAux = GerAcessos.LTotAcessos[pTotAces].ant;
    if (pTotAcesAux != NIL) // existe total anterior
      GerAcessos.LTotAcessos[pTotAcesAux].prox =
      						GerAcessos.LTotAcessos[pTotAces].prox;
    else // nao existe total anterior, este e o primeiro
	  GerAcessos.Cabeca = GerAcessos.LTotAcessos[pTotAces].prox;

    pTotAcesAux = GerAcessos.LTotAcessos[pTotAces].prox;
    if (pTotAcesAux != NIL) // existe prox total
      GerAcessos.LTotAcessos[pTotAcesAux].ant =
      						GerAcessos.LTotAcessos[pTotAces].ant;
    // fim do if

    // insere total atual na lista de dispo
    GerAcessos.LTotAcessos[pTotAces].TotAcessos = 0;
    GerAcessos.LTotAcessos[pTotAces].ant = NIL;
    GerAcessos.LTotAcessos[pTotAces].frontFilaPg =
			   GerAcessos.LTotAcessos[pTotAces].rearFilaPg = NIL;
    GerAcessos.LTotAcessos[pTotAces].prox = GerAcessos.LDispo;
    GerAcessos.LDispo = pTotAces;
  } // fim do then
  else { // existe mais de uma pg na fila
    pPgAux = ListaPgs.Pagina[pPg].antDaFila;
    if (pPgAux != NIL) // existe pagina anterior na fila - nao e a primeira
      ListaPgs.Pagina[pPgAux].proxDaFila = ListaPgs.Pagina[pPg].proxDaFila;
    else  // nao existe pagina anterior na fila - e a primeira
      GerAcessos.LTotAcessos[pTotAces].rearFilaPg =
      			 ListaPgs.Pagina[pPg].proxDaFila;

    pPgAux = ListaPgs.Pagina[pPg].proxDaFila;
    if (pPgAux != NIL) // existe prox pagina - nao e a ultima
      ListaPgs.Pagina[pPgAux].antDaFila = ListaPgs.Pagina[pPg].antDaFila;
    else // nao existe prox pagina - e a ultima
      GerAcessos.LTotAcessos[pTotAces].frontFilaPg =
      			 ListaPgs.Pagina[pPg].antDaFila;
    // fim do if
  } // fim do if
  ListaPgs.Pagina[pPg].proxDaFila =
  		   ListaPgs.Pagina[pPg].antDaFila = NIL;
  ListaPgs.Pagina[pPg].IdxTotAcessos = NIL;
}

//==============================================================================
// Metodo: void ColocaPgFila(unsigned int pPg, unsigned int pTotAces)
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::ColocaPgFila(unsigned int pPg,
                                       unsigned int pTotAces) {
  register unsigned int pPgAux;

  pPgAux = GerAcessos.LTotAcessos[pTotAces].rearFilaPg;
  ListaPgs.Pagina[pPgAux].antDaFila = pPg;
  ListaPgs.Pagina[pPg].proxDaFila = pPgAux;
  GerAcessos.LTotAcessos[pTotAces].rearFilaPg = pPg;
  ListaPgs.Pagina[pPg].antDaFila = NIL;
  ListaPgs.Pagina[pPg].IdxTotAcessos = pTotAces;
}

//==============================================================================
// Metodo: void NovoTotal(unsigned int pPg, unsigned int pTotAces)
//==============================================================================
template <class TpObjCntin>
void TNSCache<TpObjCntin>::NovoTotal(unsigned int pPg, unsigned int &pTotAces) {
  register unsigned int pTotAcesNovo;
  register unsigned int pTotAcesProx;

  // obtem um TotAcess dispo
  pTotAcesNovo = GerAcessos.LDispo;
  GerAcessos.LDispo = GerAcessos.LTotAcessos[pTotAcesNovo].prox;

  // insere novo TotAcess na lista
  GerAcessos.LTotAcessos[pTotAcesNovo].TotAcessos = 1 +
  			 GerAcessos.LTotAcessos[pTotAces].TotAcessos;
  GerAcessos.LTotAcessos[pTotAcesNovo].ant  = pTotAces;
  pTotAcesProx = GerAcessos.LTotAcessos[pTotAcesNovo].prox =
				 GerAcessos.LTotAcessos[pTotAces].prox;
  GerAcessos.LTotAcessos[pTotAces].prox = pTotAcesNovo;
  if (pTotAcesProx != NIL) // o total atual nao era o ultimo
    GerAcessos.LTotAcessos[pTotAcesProx].ant = pTotAcesNovo;
  // fim do if

  // insere pPg no novo total
  GerAcessos.LTotAcessos[pTotAcesNovo].frontFilaPg =
  			 GerAcessos.LTotAcessos[pTotAcesNovo].rearFilaPg = pPg;
  ListaPgs.Pagina[pPg].proxDaFila =
		   ListaPgs.Pagina[pPg].antDaFila = NIL;
  ListaPgs.Pagina[pPg].IdxTotAcessos = pTotAcesNovo;

  pTotAces = pTotAcesNovo; // retorna o novo pTotAces
}

#ifdef __CACHECHECK
//==============================================================================
// Metodo:
//==============================================================================
template <class TpObjCntin>
TNSCache<TpObjCntin>::checkStructure() {
  int pBlocoE, // entrada no vetor pBloco
      blocoE, // entrada no vetor TME.Bloco[]
      paginasChecadas = 0, // contador de paginas checadas
      blocosChecados = 0; // contador de blocos checados

  checkCounter++;

  // ===========================================================================
  // confere TME
  // ===========================================================================
  // Verificar se o bloco que esta ativado (diferente de NIL) esta realmente
  // sendo usado e se a pagina referente a cada entrada nesse bloco esta
  // realmente apontando para ele. Alem disso, cada pagina checada sera
  // contada, se o numero de paginas checadas for diferente de N_PAGINAS, erro.
  // Em seguida, a lista de blocos disponiveis sera checada. O numero de blocos
  // na lista de disponiveis somado ao numero de blocos ativos tem que ser igual
  // a N_PAGINAS. Nao ha a necessidade de contar o numero de entradas inativas
  // no vetor pBlocos, pois elas nao estao listadas e, se ela estiver ativada
  // e apontando para um bloco invalido, a checagem da pagina ira detectar isso.

  for (pBlocoE = 0; pBlocoE < TME.TotBlocos; pBlocoE++) {
    if (TME.pBloco[pBlocoE] != NIL) {
      int pPagina, pgChecadaLocal = 0;

      blocoE = TME.pBloco[pBlocoE]; // entrada no vetor TME.Bloco
      blocosChecados++;
      if (!TME.Bloco[blocoE].RgsUsados) { // esta ativo mas nao aponta nenhuma
                                          //  pagina
        printf("Erro em TME.Bloco[blocoE].RgsUsados."
               " checkcounter = %ld\n", checkCounter);
        exit(1);
      }

      for (pPagina = 0; pPagina < RGS_POR_BLOCO; pPagina++) {
          int pg = TME.Bloco[blocoE].pPagina[pPagina];
        if (pg != NIL) {
          pgChecadaLocal++;
          paginasChecadas++;

          if((pblocoE != (ListaPgs.Pagina[pg].IdRegFisico >> F_ROTACAO)) {
            printf("Erro em ListaPgs.Pagina[pg].IdRegFisico >> F_ROTACAO."
                   " checkcounter = %ld\n", checkCounter);
            exit(1);
          }
          if((blocoE != (ListaPgs.Pagina[pg].IdRegFisico -
                         (pblocoE << F_ROTACAO)) {
            printf("Erro em ListaPgs.Pagina[pg].IdRegFisico - (pblocoE << "
                         "F_ROTACAO. checkcounter = %ld\n", checkCounter);
            exit(1);
          }
        }
      } // garantido que todas as paginas apontam para os blocos que apontam
        // para elas.
      if (TME.Bloco[blocoE].RgsUsados != pgChecadaLocal) {
        printf("Erro em TME.Bloco[blocoE].RgsUsados != pgChecadaLocal."
               " checkcounter = %ld\n", checkCounter);
        exit(1);
      }
    } // garantido que todas as paginas apontadas por um bloco sao validas,
      // que elas apontam para eles mesmos, e que o numero de paginas apontas
      // por eles eh o numero indicado em RgsUsados
  } // garantido que os blocos ativos apontam alguma(s) pagina(s), que
    // esta(s) pagina(s) aponta(m) para o mesmo bloco que a(s) aponta(m), e
    // que o suposto numero de paginas apontadas (RgsUsados) esta correto.
  // checando TME.LBlocoDispo
  blocoE = TME.LBlocoDispo;
  while (blocoE != NIL) {
    blocosChecados++;
    blocoE = TME.Bloco[blocoE].pPagina[0];
  }
  if (blocosChecados != N_PAGINAS) {
    printf("Erro em blocosChecados != N_PAGINAS."
           " checkcounter = %ld\n", checkCounter);
    exit(1);
  } // garantido que todos os blocos que nao estao ativos estao na lista de
    // disponiveis. Por consequencia, a lista de disponiveis esta corretamente
    // encadeada

}
#endif

/*
//==============================================================================
// Metodo:
//==============================================================================
template <class TpObjCntin>
TNSCache<TpObjCntin>:: {
}
*/

//------------------------------------------------------------------------------
// Exporting templates
//------------------------------------------------------------------------------
#include "TBuffer.h"

// ---- C++ Builder || Visual C++ ----
#if defined(__BORLANDC__) || defined(_MSC_VER)
  template class __export TNSCache<TBuffer>;
#endif

// ---- GNU GCC ----
#ifdef __GNUC__
  template class TNSCache<TBuffer>;
#endif
//------------------------------------------------------------------------------

