//==============================================================================
//  Projeto: SIRIUS
//  Arquivo: TNSCache.H
//  Classe: TNSCache
//  Autor: Roberto Figueira Santos Filho
//  Orientador: Caetano Traina Junior
//  Data Inicial: 10/06/1997
//  Versao Atual: 0.0
//
//  Descricao:
//	  Este arquivo apresenta o cabecalho da classe TNSCache - Gerenciador de
//  Memoria Cache, o qual sera usado no Gerenciador do Arquivo Fisico do SIRIUS.
//
//	Observacoes:
//	  - Para facilitar o entendimento das observacoes abaixo, fica entendido
//		  que um objeto (ou classe) que possui como um de seus atributos uma
//		  instancia de um outro objeto sera chamado de continente (o que
//		  contem), e o a instancia do outro de contido.
//
//	  Um objeto da classe GFCache tem que ser declarado dentro de outro objeto,
//	nunca como objeto global, nem como variavel dentro de uma funcao e nem
//	como parametro.
//    Para a utilizacao de um objeto da classe GFCache, na sua declaracao
//	deve ser definida como template a classe continente do mesmo.
//	  Quando da instanciacao de um objeto da classe GFCache, devem ser passados
//	como parametro um ponteiro para a classe continente, ponteiros para os
//  metodos desta classe que implementam a leitura e escrita do dado a ser
//  "cacheado", o numero de registros que a base tem neste instante, o fator de
//  rotacao de bits, o numero de paginas que este cache deve ter e o tamanho
//  em bytes do dado a ser "cacheado".
//	  Esses parametros sao necessarios para tornar o cache generico.
//	  O fator de rotacao corresponde ao numero de rotacoes de bits que devem
//  ser feitas para identificar o bloco em que o registro se encontra. Supondo
//  o cache com 16 registros por bloco, temos que, dado um numero NRg de
//  registro, o bloco que ele ocupa e NRg div 16, ou NRg rotacionado de 4 bits
//  para a direita (Nreg >> 4). O fator de rotacao aqui e 4, pois 2^4 = 16.
//	  Estes metodos devem ter como parametro um ponteiro para char e devem ter
//	void como valor de retorno.
//	  Esta classe so pode ser instanciada dinamicamente - isso se da devido
//	ao fato da necessidade de o objeto continente ter que passar um ponteiro
//	para ele mesmo (this).
//
//	Exemplo:
//		// classe MyEx que usa um objeto da classe TNSCache
//	class MyEx {
//	  TNSCache<MyEx> *cache; //definindo MyEx como template do TNSCache
//    void leArq(unsigned long, unsigned char *)
//    void escreveArq(unsigned long, unsigned char *)
//	  MyEx() {	// construtor de MyEx
//	    // aqui o objeto cache deve ser instanciado passando this e os metodos
//	    // de leitura e escrita da classe continente (MyEx)
//		cache = new TNSCache<MyEx>(this, &MyEx::leArq, &MyEx::escreveArq, NRegs,
//								  FRot, NPgs, TamBuf);
//	  }
//	}
//
//    Importante -> se o objeto da classe continente criar um novo registro,
//  o metodo GMC_NovoNRgs(unsigned long reg, void **) deve ser chamado para que sejam
//  atualizados os controles do cache e tornar a pagina deste registro ativa.
//  Este metodo so deve ser chamado apos a criacao do registro fisico na base,
//  mesmo que em branco, e terem sido feitas as alteracoes na mesma para
//  refletir sua existencia. Fica por conta do usuario qualquer erro de
//  concistencia.
//
//	Modificacoes:
//
//==============================================================================

#ifndef __CACHE
#define __CACHE

//#define __CACHECHECK

//---------------- DEFINICAO DA CLASSE TNSCache -----------------

template <class TpObjCntin>
class TNSCache {
  public:
	//---- Prototipos ----
    TNSCache(TpObjCntin *, void (TpObjCntin::*)(long, void *),
    		void (TpObjCntin::*)(long, void *),
            unsigned long, unsigned char, unsigned int, unsigned long);
    ~TNSCache();
    void readBuffer(unsigned long, void **);
    void writeBuffer() {
      ListaPgs.Pagina[ListaPgs.PgEmUso].Escrita = 1;
    } // fim do metodo writeBuffer
    void newBuffer(unsigned long, void **);
    void flushCache();

  private:
    #define TxCrescBlocos 51
    #define NIL 0xFFFFU
	//---- Tipos ----
	//-- estrutura para mapeamento de enderecos de registros
    //   para paginas em memoria
    typedef struct {
      unsigned int *pPagina; // vetor com RGS_POR_BLOCO posicoes
      unsigned char RgsUsados;
    } tBloco;

	typedef struct { // tabela de mapeamento de enderecos
	  unsigned long TotBlocos;
      unsigned int *pBloco; // vetor com TotBlocos posicoes
      tBloco *Bloco; // vetor com N_PAGINAS posicoes
      unsigned int LBlocoDispo;
    } tTME;

	//-- estrutura para manter a lista de paginas do cache
    typedef struct {
      unsigned int proxDaFila;
      unsigned int antDaFila;
      unsigned int IdxTotAcessos;
      unsigned long IdRegFisico;
      unsigned char  Escrita;
      unsigned char *Buffer;
    } tPagina;

    typedef struct {
      tPagina *Pagina; // vetor com N_PAGINAS posicoes
      unsigned int LPgDispo;
      unsigned int PgEmUso;
    } tListaPgs;

	//-- estrutura para manter o Gerenciador de Acessos aas paginas do Cache
    typedef struct {
	  unsigned int TotAcessos; // indica o total de acessos deste no da lista de totais
	  unsigned int prox; // proximo no da lista de totais de acessos
	  unsigned int ant; // no anterior da lista de totais de acessos
      unsigned int frontFilaPg; // aponta para a pg da fila de pgs referenciada a mais
      			   // tempo. As paginas sao retiradas daqui.
      unsigned int rearFilaPg; // aponta para a pg da fila de pgs referenciada a menos
      			  // tempo. Novas paginas sao inseridas aqui.
    } tLTotAcessos;

    typedef struct {
	  tLTotAcessos *LTotAcessos; // vetor com os totais de acessos, possui
      						   // N_PAGINAS posicoes
      unsigned int Cabeca; // aponta para o de menor total - primeiro da lista
      unsigned int LDispo;
    } tGerAcessos;

	//---- Variaveis ----
	unsigned char F_ROTACAO;
    unsigned char RGS_POR_BLOCO;
    unsigned int N_PAGINAS;

	tTME TME;
    tListaPgs ListaPgs;
	tGerAcessos GerAcessos;
	TpObjCntin *ObjContinente; // objeto continente de um objeto desta classe
	//---- Prototipos ----
    void (TpObjCntin::*leC)(long, void *);
    void (TpObjCntin::*escreveC)(long, void *);

    void IncPagina(); // incrementa o numero de acessos da pagina corrente
    unsigned int ObtemPagina();
	void TiraPgFila(unsigned int, unsigned int);
	void ColocaPgFila(unsigned int, unsigned int);
	void NovoTotal(unsigned int, unsigned int &);
    #ifdef __CACHECHECK
      long checkCounter; // deve ser usado para identificar o momento em que o
                         // erro ocorreu. Ao inves de ficar percorrendo o cache
                         // ate o erro (e isso pode implicar em MUITOS acessos),
                         // este contador eh incrementado e impresso quando um
                         // um erro na estrutura eh encontrado. Dai, basta o
                         // programador gerar um break quando o contador chegar
                         // no valor do erro novamente.
      void checkStructure(); // checa a estrutura do cache em busca de erros
                             // util para debugar
    #endif
//	void InicPar();
};
#endif

/*
//==============================================================================
Resumo das variaveis, parametros e constantes

--- Constantes
TxCrescBlocos 51
NIL 0xFFFF

--- Parametros constantes
F_ROTACAO
RGS_POR_BLOCO
N_PAGINAS

--- Variaveis
// inicializando TME
TME.TotBlocos
TME.pBloco[]
TME.Bloco[].pPagina[]
TME.Bloco[].RgsUsados
TME.LBlocoDispo

// inicializando ListaPgs
ListaPgs.Pagina[].proxDaFila
ListaPgs.Pagina[].antDaFila
ListaPgs.Pagina[].IdxTotAcessos
ListaPgs.Pagina[].IdRegFisico
ListaPgs.Pagina[].Escrita
ListaPgs.Pagina[].Buffer
ListaPgs.LPgDispo
ListaPgs.PgEmUso

// inicializando GerAcessos
GerAcessos.LTotAcessos[].TotAcessos
GerAcessos.LTotAcessos[].prox
GerAcessos.LTotAcessos[].ant
GerAcessos.LTotAcessos[].frontFilaPg
GerAcessos.LTotAcessos[].rearFilaPg
GerAcessos.Cabeca
GerAcessos.LDispo

--- Outros parametros
ObjContinente
(ObjContinente->*leC)(IdRegFisico, Buffer);
(ObjContinente->*escreveC)(IdRegFisico, Buffer);
//==============================================================================
*/
